/*
 * Copyright (C) 2001-2011 OSS Nokalva.  All rights reserved.
 *
 */

/*
 * THIS FILE IS PROPRIETARY MATERIAL OF OSS NOKALVA, INC. AND MAY BE USED
 * ONLY BY DIRECT LICENSEES OF OSS NOKALVA, INC. THIS FILE MAY NOT BE
 * DISTRIBUTED.
 */

/*
 * In this example the BER-encoder is called to create an encoded PDU.
 * Then the OSS IAAPI is used to parse the raw BER-encoded PDU for both
 * DEFINITE- and INDEFINITE-length cases. After a given tag sequence is
 * found, the decoder is called to decode the found component as a
 * stand-alone encoded PDU.
 */



#include "iaapi.h"
#include "bcas.h"


typedef struct {
    int curOccurrence;		/* stores current occurrence of tag sequence */
    int searchOccurrence;	/* ocurrence of the tag sequence searched */
    int numberOfTags;		/* number of tags in the tag sequence searched */
    int index;			/* index into the tag sequence searched */
    short *tag;			/* pointer to the tag sequence searched */
} Tags;


static void parseBEREncoding(OssGlobal *world, unsigned char *curEnc,
	long *remainingLength, long *offset, long offsetSaved, Tags *tags);


int main(void)
{

    OssGlobal w, *world = &w;
    long length, offset;
    int rc = 0, pdu_num;
    OssBuf encodedPDU = {0};
    void *decodedPDU, *buf;
    Tags tag;
    unsigned short tags[] = {10, 1, 137, 138};


    rc = ossinit(world, bcas);
    if (rc) {
	printf("Initialization error = %d\n", rc);
	return 1;
    }

    ossPrint(world, "Running for DEFINITE-length encoding...\n\n");

    ossSetFlags(world, DEBUGPDU|DEFINITE);

    ossPrint(world, "\tUnencoded PDU...\n\n");
    ossPrintPDU(world, BBCard_PDU, &myCard);

    ossPrint(world, "The encoder's trace messages...\n\n");

    encodedPDU.value = NULL;
    encodedPDU.length = 0;

    if (ossEncode(world, BBCard_PDU, &myCard, &encodedPDU)) {
        ossPrint(world, "%s\n", ossGetErrMsg(world));
	ossterm(world);
        return 1;
    }

    buf = (char *)encodedPDU.value;

    length = encodedPDU.length;
    offset = 0;

	/*
	 * Initialize an instance of the Tags structure to find the beginning
	 * of the ENUMERATED field in the DEFINITE-length encoded PDU.
	 */
    tag.tag = tags;
    tag.index = 0;
    tag.searchOccurrence = 1;
    pdu_num = Enum_PDU;
    tag.curOccurrence = 1;
    tag.numberOfTags = 1;

    parseBEREncoding(world, encodedPDU.value, &length, &offset, 0, &tag);


    if (offset < encodedPDU.length) {
	encodedPDU.value += offset;
	decodedPDU = NULL;

	ossPrint(world, "\nThe decoder's trace messages...\n\n");

	if (ossDecode(world, &pdu_num, &encodedPDU, &decodedPDU)) {
	    ossPrint(world, "%s\n", ossGetErrMsg(world));
	    ossterm(world);
	    return 1;
	}
	else {
	    ossPrint(world, "\n\tDecoded PDU...\n\n");
	    ossPrintPDU(world, pdu_num, decodedPDU);
	    ossFreePDU(world, pdu_num, decodedPDU);
	}
    }

    encodedPDU.value = buf;
    length = encodedPDU.length;
    offset = 0;

	/*
	 * Initialize an instance of the Tags structure to find the beginning
	 * of the REAL field in the DEFINITE-length encoded PDU.
	 */
    tag.tag = &tags[1];
    tag.index = 0;
    tag.searchOccurrence = 2;
    pdu_num = Real_PDU;
    tag.curOccurrence = 1;
    tag.numberOfTags = 3;

    parseBEREncoding(world, encodedPDU.value, &length, &offset, 0, &tag);


    if (offset < encodedPDU.length) {
	encodedPDU.value += offset;
	decodedPDU = NULL;

	ossPrint(world, "\nThe decoder's trace messages...\n\n");

	if (ossDecode(world, &pdu_num, &encodedPDU, &decodedPDU)) {
	    ossPrint(world, "%s\n", ossGetErrMsg(world));
	    ossterm(world);
	    return 1;
	}
	else {
	    ossPrint(world, "\n\tDecoded PDU...\n\n");
	    ossPrintPDU(world, pdu_num, decodedPDU);
	    ossFreePDU(world, pdu_num, decodedPDU);
	}
    }

    ossFreeBuf(world, buf);


    ossPrint(world, "\n\nRunning for INDEFINITE-length encoding...\n\n");

    ossSetFlags(world, DEBUGPDU|INDEFINITE);

    ossPrint(world, "\tUnencoded PDU...\n\n");
    ossPrintPDU(world, BBCard_PDU, &myCard);

    ossPrint(world, "The encoder's trace messages...\n\n");

    encodedPDU.value = NULL;
    encodedPDU.length = 0;

    if (ossEncode(world, BBCard_PDU, &myCard, &encodedPDU)) {
        ossPrint(world, "%s\n", ossGetErrMsg(world));
	ossterm(world);
        return 1;
    }
    ossPrintHex(world, encodedPDU.value, encodedPDU.length);
    buf = (char *)encodedPDU.value;

    length = encodedPDU.length;
    offset = 0;

	/*
	 * Initialize an instance of the Tags structure to find the beginning
	 * of the ENUMERATED field in the INDEFINITE-length encoded PDU.
	 */
    tag.tag = tags;
    tag.index = 0;
    tag.searchOccurrence = 1;
    pdu_num = Enum_PDU;
    tag.curOccurrence = 1;
    tag.numberOfTags = 1;

    parseBEREncoding(world, encodedPDU.value, &length, &offset, 0, &tag);


    if (offset < encodedPDU.length) {
	encodedPDU.value += offset;
	decodedPDU = NULL;

	ossPrint(world, "\nThe decoder's trace messages...\n\n");

	if (ossDecode(world, &pdu_num, &encodedPDU, &decodedPDU)) {
	    ossPrint(world, "%s\n", ossGetErrMsg(world));
	    rc = 1;
	}
	else {
	    ossPrint(world, "\n\tDecoded PDU...\n\n");
	    ossPrintPDU(world, pdu_num, decodedPDU);
	    ossFreePDU(world, pdu_num, decodedPDU);
	}
    }

    encodedPDU.value = buf;
    length = encodedPDU.length;
    offset = 0;

	/*
	 * Initialize an instance of the Tags structure to find the beginning
	 * of the REAL field in the INDEFINITE-length encoded PDU.
	 */
    tag.tag = &tags[1];
    tag.index = 0;
    tag.searchOccurrence = 2;
    pdu_num = Real_PDU;
    tag.curOccurrence = 1;
    tag.numberOfTags = 3;

    parseBEREncoding(world, encodedPDU.value, &length, &offset, 0, &tag);


    if (offset < encodedPDU.length) {
	encodedPDU.value += offset;
	decodedPDU = NULL;

	ossPrint(world, "\nThe decoder's trace messages...\n\n");

	if (ossDecode(world, &pdu_num, &encodedPDU, &decodedPDU)) {
	    ossPrint(world, "%s\n", ossGetErrMsg(world));
	    rc = 1;
	}
	else {
	    ossPrint(world, "\n\tDecoded PDU...\n\n");
	    ossPrintPDU(world, pdu_num, decodedPDU);
	    ossFreePDU(world, pdu_num, decodedPDU);
	}
    }

    ossFreeBuf(world, buf);
    ossterm(world);
    return rc;
}


/*
 * Function:            parseBEREncoding() parses raw DEFINITE & INDEFINITE
 *                      BER-encodings looking for tag-length pairs.
 *
 * Parameters:
 *
 *   "world"        ... pointer to the OSS's Global structure
 *   "encoding"     ... pointer to encoded PDU
 *   "remainingLength". remaning length of the encoding
 *   "offset"       ... offset to the tag sequence searched
 *   "offsetSaved"  ... variable used to safe the offset to the found tag
 *                      sequence and later to return it in "offset"
 *   "tags"         ... pointer to the Tags structure used to pass tag-related
 *                      information for the tag sequence searched
 *
 * Return value:        none
 */
static void parseBEREncoding(OssGlobal *world,
	unsigned char *encoding, long *remainingLength,
			long *offset, long offsetSaved, Tags *tags)
{

    unsigned char	*curEnc = encoding;
    unsigned char	*nestedEnc, *nextEnc;
    long		nestedRemainingLength = *remainingLength;
    long		curRemainingLength, nestedOffset;
    int			rc = 0;
    LONG_LONG		tag;


    do {
	curRemainingLength = *remainingLength;
	nestedRemainingLength = *remainingLength;
	nestedEnc = NULL;
	nestedOffset = 0;
	nextEnc = NULL;
	/*
	 * Try to get the nested encoding.  If the current encoding contains
	 * nested encodings, parseBEREncoding() is called recursively.
	 */
	if (!rc && (nestedEnc = ossGetNestedBEREncoding(world, curEnc,
			&nestedRemainingLength)) && nestedRemainingLength) {
		/*
		 * Update "offset" to keep an offset of the first nested
		 * encoding.
		 */
	    *offset += nestedEnc - curEnc;
	    nestedOffset = nestedRemainingLength + (nestedEnc - curEnc);
	    tag = ossEncodingASN1Tag(world, nestedEnc, nestedRemainingLength, 1);
	    if (tags->tag[tags->index++] == tag && tags->index == tags->numberOfTags) {
		if (tags->curOccurrence++ == tags->searchOccurrence) {
		    if (offsetSaved)
			*offset = offsetSaved;
		    *remainingLength = -1;
		    break;
		}
		else
		    tags->index = 0;
	    }
	    else if (tags->tag[tags->index - 1] != tag)
		tags->index = 0;
	    parseBEREncoding(world, nestedEnc, &nestedRemainingLength,
						offset, offsetSaved, tags);
	    if (nestedRemainingLength == -1) {
		*remainingLength = -1;
		break;
	    }
	}
		/*
		 * Get the next encoding.
		 */
	if (!rc)
	    nextEnc = ossGetNextBEREncoding(world, curEnc, remainingLength);
		/*
		 * Update and safe "offset" to keep an offset of the next
		 * encoding if the current encoding does not have nested
		 * encodings. If it has nested encodings "offset" is updated
		 * in the recursive call to parseBEREncoding().
		 */
	if (!nestedEnc) 
	    *offset += nextEnc ? nextEnc - curEnc : curRemainingLength;
	else if (nestedOffset && nestedOffset < (curRemainingLength - *remainingLength))
	    /* Skip EOCs from parent encodings in case of indefinite lengths. */
	    *offset += curRemainingLength - *remainingLength - nestedOffset;
	offsetSaved = *offset;
	tag = ossEncodingASN1Tag(world, nextEnc, *remainingLength, 1);
	if (tags->tag[tags->index++] == tag && tags->index == tags->numberOfTags) {
	    if (tags->curOccurrence++ == tags->searchOccurrence) {
		*remainingLength = -1;
		break;
	    }
	    else
		tags->index = 0;
	}
	else if (tags->tag[tags->index - 1] != tag)
	    tags->index = 0;
	curEnc = nextEnc;
    } while (nextEnc && *remainingLength);
}


